"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.nftValuetoXrpl = exports.xrplValueToNft = exports.currencyCodeFormat = exports.normalizeNodes = exports.normalizeNode = exports.dropsToXRP = void 0;
exports.dropsToXRP = (drops) => {
    return drops.dividedBy(1000000);
};
exports.normalizeNode = (affectedNode) => {
    const diffType = Object.keys(affectedNode)[0];
    const node = affectedNode[diffType];
    return Object.assign({}, Object.assign(Object.assign({}, node), { diffType, entryType: node.LedgerEntryType, ledgerIndex: node.LedgerIndex, newFields: node.NewFields || {}, finalFields: node.FinalFields || {}, previousFields: node.PreviousFields || {} }));
};
exports.normalizeNodes = (metadata) => {
    if (!metadata.AffectedNodes) {
        return [];
    }
    return metadata.AffectedNodes.map(exports.normalizeNode);
};
// Supports regular 3 char currency code
// Supports HEX currency code
// Supports HEX currency code based on XLS15d
exports.currencyCodeFormat = (string, maxLength = 12, nativeAsset = 'XRP') => {
    if (string.length === 3 && string.trim().toLowerCase() !== nativeAsset.toLowerCase()) {
        // Normal currency code
        return string.trim();
    }
    if (string.match(/^[a-fA-F0-9]{40}$/) && !isNaN(parseInt(string, 16))) {
        // HEX currency code
        const hex = string.toString().replace(/(00)+$/g, '');
        if (hex.startsWith('02')) {
            const xlf15d = Buffer.from(hex, 'hex').slice(8).toString('utf-8').slice(0, maxLength).trim();
            if (xlf15d.match(/[a-zA-Z0-9]{3,}/) && xlf15d.toLowerCase() !== nativeAsset.toLowerCase()) {
                return xlf15d;
            }
        }
        const decodedHex = Buffer.from(hex, 'hex').toString('utf-8').slice(0, maxLength).trim();
        if (decodedHex.match(/[a-zA-Z0-9]{3,}/) && decodedHex.toLowerCase() !== nativeAsset.toLowerCase()) {
            return decodedHex;
        }
    }
    return '???';
};
// XLS-14d Sample implementation,
//   https://gist.github.com/WietseWind/5ffbf67cd982a7e9bd8f0ded52e60fe3
//   https://hash.xrp.fans/8F3CE0481EF31A1BE44AD7D744D286B0F440780CD0056951948F93A803D47F8B
exports.xrplValueToNft = (value) => {
    const data = String(Number(value)).split(/e/i);
    const finish = (returnValue) => {
        const unsignedReturnValue = returnValue.replace(/^\-/, '');
        if (unsignedReturnValue.length > 83) {
            // Too many tokens to be NFT-like as per XLS14d proposal
            return false;
        }
        if (data.length > 1 && unsignedReturnValue.slice(0, 2) === '0.' && Number(data[1]) < -70) {
            // Positive below zero amount, could be NFT
            return (sign === '-' ? -1 : 1) * Number((unsignedReturnValue.slice(2) + '0'.repeat(83 - unsignedReturnValue.length))
                .replace(/^0+/, ''));
        }
        return false;
    };
    if (data.length === 1) {
        // Regular (non-exponent)
        return false;
    }
    let z = '';
    const sign = Number(value) < 0 ? '-' : '';
    const str = data[0].replace('.', '');
    let mag = Number(data[1]) + 1;
    if (mag < 0) {
        z = sign + '0.';
        while (mag++) {
            z += '0';
        }
        return finish(z + str.replace(/^\-/, ''));
    }
    mag -= str.length;
    while (mag--) {
        z += '0';
    }
    return finish(str + z);
};
exports.nftValuetoXrpl = (value, accountBalance) => {
    const unsignedValue = String(value).replace(/^-/, '');
    const sign = unsignedValue.length < String(value).length ? '-' : '';
    // accountBalance: xrpl string notation, optional, if intention to force NFT check
    if (typeof accountBalance !== 'undefined' && exports.xrplValueToNft(accountBalance) === false) {
        throw new Error('Source balance is not NFT-like');
    }
    if (!unsignedValue.match(/^[0-9]+$/)) {
        throw new Error('Only non-float & non-scientific notation values accepted');
    }
    return sign + '0.' + '0'.repeat(81 - unsignedValue.length) + unsignedValue;
};
